"""Contains decorator functions and functions for interacting with global data.
"""
from functools import wraps

import requests
from robin_stocks.robinhood.globals import LOGGED_IN, OUTPUT, SESSION


def set_login_state(logged_in):
    """Sets the login state"""
    global LOGGED_IN
    LOGGED_IN = logged_in

def set_output(output):
    """Sets the global output stream"""
    global OUTPUT
    OUTPUT = output
    
def get_output():
    """Gets the current global output stream"""
    global OUTPUT
    return OUTPUT

def login_required(func):
    """A decorator for indicating which methods require the user to be logged
       in."""
    @wraps(func)
    def login_wrapper(*args, **kwargs):
        global LOGGED_IN
        if not LOGGED_IN:
            raise Exception('{} can only be called when logged in'.format(
                func.__name__))
        return(func(*args, **kwargs))
    return(login_wrapper)


def convert_none_to_string(func):
    """A decorator for converting a None Type into a blank string"""
    @wraps(func)
    def string_wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if result:
            return(result)
        else:
            return("")
    return(string_wrapper)


def id_for_stock(symbol):
    """Takes a stock ticker and returns the instrument id associated with the stock.

    :param symbol: The symbol to get the id for.
    :type symbol: str
    :returns:  A string that represents the stocks instrument id.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return(None)

    url = 'https://api.robinhood.com/instruments/'
    payload = {'symbol': symbol}
    data = request_get(url, 'indexzero', payload)

    return(filter_data(data, 'id'))


def id_for_chain(symbol):
    """Takes a stock ticker and returns the chain id associated with a stocks option.

    :param symbol: The symbol to get the id for.
    :type symbol: str
    :returns:  A string that represents the stocks options chain id.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return(None)

    url = 'https://api.robinhood.com/instruments/'

    payload = {'symbol': symbol}
    data = request_get(url, 'indexzero', payload)

    if data:
        return(data['tradable_chain_id'])
    else:
        return(data)


def id_for_group(symbol):
    """Takes a stock ticker and returns the id associated with the group.

    :param symbol: The symbol to get the id for.
    :type symbol: str
    :returns:  A string that represents the stocks group id.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return(None)

    url = 'https://api.robinhood.com/options/chains/{0}/'.format(
        id_for_chain(symbol))
    data = request_get(url)
    return(data['underlying_instruments'][0]['id'])


def id_for_option(symbol, expirationDate, strike, optionType):
    """Returns the id associated with a specific option order.

    :param symbol: The symbol to get the id for.
    :type symbol: str
    :param expirationData: The expiration date as YYYY-MM-DD
    :type expirationData: str
    :param strike: The strike price.
    :type strike: str
    :param optionType: Either call or put.
    :type optionType: str
    :returns:  A string that represents the stocks option id.

    """ 
    symbol = symbol.upper()
    chain_id = id_for_chain(symbol)
    payload = {
        'chain_id': chain_id,
        'expiration_dates': expirationDate,
        'strike_price': strike,
        'type': optionType,
        'state': 'active'
    }
    url = 'https://api.robinhood.com/options/instruments/'
    data = request_get(url, 'pagination', payload)

    listOfOptions = [item for item in data if item["expiration_date"] == expirationDate]
    if (len(listOfOptions) == 0):
        print('Getting the option ID failed. Perhaps the expiration date is wrong format, or the strike price is wrong.', file=get_output())
        return(None)

    return(listOfOptions[0]['id'])


def round_price(price):
    """Takes a price and rounds it to an appropriate decimal place that Robinhood will accept.

    :param price: The input price to round.
    :type price: float or int
    :returns: The rounded price as a float.

    """
    price = float(price)
    if price <= 1e-2:
        returnPrice = round(price, 6)
    elif price < 1e0:
        returnPrice = round(price, 4)
    else:
        returnPrice = round(price, 2)

    return returnPrice


def filter_data(data, info):
    """Takes the data and extracts the value for the keyword that matches info.

    :param data: The data returned by request_get.
    :type data: dict or list
    :param info: The keyword to filter from the data.
    :type info: str
    :returns:  A list or string with the values that correspond to the info keyword.

    """
    if (data == None):
        return(data)
    elif (data == [None]):
        return([])
    elif (type(data) == list):
        if (len(data) == 0):
            return([])
        compareDict = data[0]
        noneType = []
    elif (type(data) == dict):
        compareDict = data
        noneType = None

    if info is not None:
        if info in compareDict and type(data) == list:
            return([x[info] for x in data])
        elif info in compareDict and type(data) == dict:
            return(data[info])
        else:
            print(error_argument_not_key_in_dictionary(info), file=get_output())
            return(noneType)
    else:
        return(data)


def inputs_to_set(inputSymbols):
    """Takes in the parameters passed to *args and puts them in a set and a list.
    The set will make sure there are no duplicates, and then the list will keep
    the original order of the input.

    :param inputSymbols: A list, dict, or tuple of stock tickers.
    :type inputSymbols: list or dict or tuple or str
    :returns:  A list of strings that have been capitalized and stripped of white space.

    """

    symbols_list = []
    symbols_set = set()

    def add_symbol(symbol):
        symbol = symbol.upper().strip()
        if symbol not in symbols_set:
            symbols_set.add(symbol)
            symbols_list.append(symbol)

    if type(inputSymbols) is str:
        add_symbol(inputSymbols)
    elif type(inputSymbols) is list or type(inputSymbols) is tuple or type(inputSymbols) is set:
        inputSymbols = [comp for comp in inputSymbols if type(comp) is str]
        for item in inputSymbols:
            add_symbol(item)

    return(symbols_list)


def request_document(url, payload=None):
    """Using a document url, makes a get request and returnes the session data.

    :param url: The url to send a get request to.
    :type url: str
    :returns: Returns the session.get() data as opppose to session.get().json() data.

    """ 
    try:
        res = SESSION.get(url, params=payload)
        res.raise_for_status()
    except requests.exceptions.HTTPError as message:
        print(message, file=get_output())
        return(None)

    return(res)


def request_get(url, dataType='regular', payload=None, jsonify_data=True):
    """For a given url and payload, makes a get request and returns the data.

    :param url: The url to send a get request to.
    :type url: str
    :param dataType: Determines how to filter the data. 'regular' returns the unfiltered data. \
    'results' will return data['results']. 'pagination' will return data['results'] and append it with any \
    data that is in data['next']. 'indexzero' will return data['results'][0].
    :type dataType: Optional[str]
    :param payload: Dictionary of parameters to pass to the url. Will append the requests url as url/?key1=value1&key2=value2.
    :type payload: Optional[dict]
    :param jsonify_data: If this is true, will return requests.post().json(), otherwise will return response from requests.post().
    :type jsonify_data: bool
    :returns: Returns the data from the get request. If jsonify_data=True and requests returns an http code other than <200> \
    then either '[None]' or 'None' will be returned based on what the dataType parameter was set as.

    """
    if (dataType == 'results' or dataType == 'pagination'):
        data = [None]
    else:
        data = None
    res = None
    if jsonify_data:
        try:
            res = SESSION.get(url, params=payload)
            res.raise_for_status()
            data = res.json()
        except (requests.exceptions.HTTPError, AttributeError) as message:
            print(message, file=get_output())
            return(data)
    else:
        res = SESSION.get(url, params=payload)
        return(res)
    # Only continue to filter data if jsonify_data=True, and Session.get returned status code <200>.
    if (dataType == 'results'):
        try:
            data = data['results']
        except KeyError as message:
            print("{0} is not a key in the dictionary".format(message), file=get_output())
            return([None])
    elif (dataType == 'pagination'):
        counter = 2
        nextData = data
        try:
            data = data['results']
        except KeyError as message:
            print("{0} is not a key in the dictionary".format(message), file=get_output())
            return([None])

        if nextData['next']:
            print('Found Additional pages.', file=get_output())
        while nextData['next']:
            try:
                res = SESSION.get(nextData['next'])
                res.raise_for_status()
                nextData = res.json()
            except:
                print('Additional pages exist but could not be loaded.', file=get_output())
                return(data)
            print('Loading page '+str(counter)+' ...', file=get_output())
            counter += 1
            for item in nextData['results']:
                data.append(item)
    elif (dataType == 'indexzero'):
        try:
            data = data['results'][0]
        except KeyError as message:
            print("{0} is not a key in the dictionary".format(message), file=get_output())
            return(None)
        except IndexError as message:
            return(None)

    return(data)


def request_post(url, payload=None, timeout=16, json=False, jsonify_data=True):
    """For a given url and payload, makes a post request and returns the response. Allows for responses other than 200.

    :param url: The url to send a post request to.
    :type url: str
    :param payload: Dictionary of parameters to pass to the url as url/?key1=value1&key2=value2.
    :type payload: Optional[dict]
    :param timeout: The time for the post to wait for a response. Should be slightly greater than multiples of 3.
    :type timeout: Optional[int]
    :param json: This will set the 'content-type' parameter of the session header to 'application/json'
    :type json: bool
    :param jsonify_data: If this is true, will return requests.post().json(), otherwise will return response from requests.post().
    :type jsonify_data: bool
    :returns: Returns the data from the post request.

    """
    data = None
    res = None
    try:
        if json:
            update_session('Content-Type', 'application/json')
            res = SESSION.post(url, json=payload, timeout=timeout)
            update_session(
                'Content-Type', 'application/x-www-form-urlencoded; charset=utf-8')
        else:
            res = SESSION.post(url, data=payload, timeout=timeout)
        if res.status_code not in [200, 201, 202, 204, 301, 302, 303, 304, 307, 400, 401, 402, 403]:
            raise Exception("Received "+ str(res.status_code))
        data = res.json()
    except Exception as message:
        print("Error in request_post: {0}".format(message), file=get_output())
    if jsonify_data:
        return(data)
    else:
        return(res)


def request_delete(url):
    """For a given url and payload, makes a delete request and returns the response.

    :param url: The url to send a delete request to.
    :type url: str
    :returns: Returns the data from the delete request.

    """
    try:
        res = SESSION.delete(url)
        res.raise_for_status()
        data = res
    except Exception as message:
        data = None
        print("Error in request_delete: {0}".format(message), file=get_output())
        
    return(data)


def update_session(key, value):
    """Updates the session header used by the requests library.

    :param key: The key value to update or add to session header.
    :type key: str
    :param value: The value that corresponds to the key.
    :type value: str
    :returns: None. Updates the session header with a value.

    """
    SESSION.headers[key] = value


def error_argument_not_key_in_dictionary(keyword):
    return('Error: The keyword "{0}" is not a key in the dictionary.'.format(keyword))


def error_ticker_does_not_exist(ticker):
    return('Warning: "{0}" is not a valid stock ticker. It is being ignored'.format(ticker))


def error_must_be_nonzero(keyword):
    return('Error: The input parameter "{0}" must be an integer larger than zero and non-negative'.format(keyword))



from robin_stocks.tda.helper import format_inputs, login_required, request_get
from robin_stocks.tda.urls import URLS


@login_required
@format_inputs
def get_accounts(options=None, jsonify=None):
    """ Gets all accounts associated with your API keys.

    :param options: Balances displayed by default, additional fields can be added here by adding positions or orders\
        As a comma separated list. Example:"positions,orders"
    :type options: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.
    """
    url = URLS.accounts()
    if options:
        payload = {
            "fields": options
        }
    else:
        payload = None
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_account(id, options=None, jsonify=None):
    """ Get account information for a specific account.

    :param id: The account id.
    :type id: str
    :param options: Balances displayed by default, additional fields can be added here by adding positions or orders\
        As a comma separated list. Example:"positions,orders"
    :type options: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.
    """
    url = URLS.account(id)
    if options:
        payload = {
            "fields": options
        }
    else:
        payload = None
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_transactions(id, type_value=None, symbol=None, start_date=None, end_date=None, jsonify=None):
    """ Get account information for a specific account.

    :param id: The account id.
    :type id: str
    :param type_value: Only transactions with the specified type will be returned. ALL, TRADE, \
        BUY_ONLY, SELL_ONLY, CASH_IN_OR_CASH_OUT, CHECKING, DIVIDEND, INTEREST, OTHER, ADVISOR_FEES
    :type type_value: Optional[str]
    param symbol: Only transactions with the specified symbol will be returned.
    :type symbol: Optional[str]
    param start_date: Only transactions after the Start Date will be returned. \
        Note: The maximum date range is one year. Valid ISO-8601 formats are :yyyy-MM-dd.
    :type start_date: Optional[str]
    param end_date: Only transactions before the End Date will be returned. \
        Note: The maximum date range is one year. Valid ISO-8601 formats are :yyyy-MM-dd.
    :type end_date: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.
    """
    url = URLS.transactions(id)
    payload = {}
    if type_value:
        payload["type"] = type_value
    if symbol:
        payload["symbol"] = symbol
    if start_date:
        payload["startDate"] = start_date
    if end_date:
        payload["endDate"] = end_date
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_transaction(account_id, transaction_id, jsonify=None):
    """ Get account information for a specific account.

    :param account_id: The account id.
    :type account_id: str
    :param transaction_id: The transaction id.
    :type transaction_id: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.
    """
    url = URLS.transaction(account_id, transaction_id)
    data, error = request_get(url, None, jsonify)
    return data, error



from robin_stocks.gemini.authentication import (generate_order_id,
                                                generate_signature)
from robin_stocks.gemini.crypto import get_price
from robin_stocks.gemini.helper import (format_inputs, login_required,
                                        request_post)
from robin_stocks.gemini.urls import URLS


@login_required
@format_inputs
def get_trades_for_crypto(ticker, limit_trades=50, timestamp=None, jsonify=None):
    """ Gets a list of all transactions for a certain crypto.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param limit_trades: The maximum number of trades to return. Default is 50, max is 500.
    :type limit_trades: Optional[int]
    :param timestamp: Only return trades on or after this timestamp. If not present, will show the most recent orders
    :type timestamp: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * price
                      * amount
                      * timestamp
                      * timestampms
                      * type
                      * aggressor
                      * fee_currency
                      * fee_amount
                      * tid
                      * order_id
                      * exchange
                      * is_auction_fill
                      * client_order_id

    """
    url = URLS.mytrades()
    payload = {
        "request": URLS.get_endpoint(url),
        "symbol": ticker,
        "limit_trades": limit_trades
    }
    if timestamp:
        payload["timestamp"] = timestamp

    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def cancel_all_session_orders(jsonify=None):
    """ Cancel all orders opened by the session.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * result
                      * details

    """
    url = URLS.cancel_session_orders()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def cancel_all_active_orders(jsonify=None):
    """ Cancel all orders for all sessions opened by the account.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * result
                      * details

    """
    url = URLS.cancel_active_orders()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def cancel_order(order_id, jsonify=None):
    """ Cancel a specific order based on ID.

    :param order_id: The id of the order. This is not the same as the client order ID.
    :type order_id: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * order_id - The order id
                      * client_order_id - An optional client-specified order id
                      * symbol - The symbol of the order
                      * exchange - Will always be "gemini"
                      * price - The price the order was issued at
                      * avg_execution_price - The average price at which this order as been executed so far. 0 if the order has not been executed at all.
                      * side - Either "buy" or "sell".
                      * type - Description of the order.
                      * options - An array containing at most one supported order execution option.
                      * timestamp - The timestamp the order was submitted. Note that for compatibility reasons, this is returned as a string. We recommend using the timestampms field instead.
                      * timestampms - The timestamp the order was submitted in milliseconds.
                      * is_live - true if the order is active on the book (has remaining quantity and has not been canceled)
                      * is_cancelled - true if the order has been canceled. Note the spelling, "cancelled" instead of "canceled". This is for compatibility reasons.
                      * reason - Populated with the reason your order was canceled, if available.
                      * was_forced - Will always be false.
                      * executed_amount - The amount of the order that has been filled.
                      * remaining_amount - The amount of the order that has not been filled.
                      * original_amount - The originally submitted amount of the order.
                      * is_hidden - Will always return false unless the order was placed with the indication-of-interest execution option.

    """
    url = URLS.cancel_order()
    payload = {
        "request": URLS.get_endpoint(url),
        "order_id": order_id
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def order_status(order_id, jsonify=None):
    """ Get the status for an order.

    :param order_id: The id of the order. This is not the same as the client order ID.
    :type order_id: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * order_id - The order id
                      * client_order_id - An optional client-specified order id
                      * symbol - The symbol of the order
                      * exchange - Will always be "gemini"
                      * price - The price the order was issued at
                      * avg_execution_price - The average price at which this order as been executed so far. 0 if the order has not been executed at all.
                      * side - Either "buy" or "sell".
                      * type - Description of the order.
                      * options - An array containing at most one supported order execution option.
                      * timestamp - The timestamp the order was submitted. Note that for compatibility reasons, this is returned as a string. We recommend using the timestampms field instead.
                      * timestampms - The timestamp the order was submitted in milliseconds.
                      * is_live - true if the order is active on the book (has remaining quantity and has not been canceled)
                      * is_cancelled - true if the order has been canceled. Note the spelling, "cancelled" instead of "canceled". This is for compatibility reasons.
                      * reason - Populated with the reason your order was canceled, if available.
                      * was_forced - Will always be false.
                      * executed_amount - The amount of the order that has been filled.
                      * remaining_amount - The amount of the order that has not been filled.
                      * original_amount - The originally submitted amount of the order.
                      * is_hidden - Will always return false unless the order was placed with the indication-of-interest execution option.

    """
    url = URLS.order_status()
    payload = {
        "request": URLS.get_endpoint(url),
        "order_id": order_id
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def active_orders(jsonify=None):
    """ Get a list of all active orders.

    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a list of dictionaries parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionaries are listed below.
    :Dictionary Keys: * order_id - The order id
                      * client_order_id - An optional client-specified order id
                      * symbol - The symbol of the order
                      * exchange - Will always be "gemini"
                      * price - The price the order was issued at
                      * avg_execution_price - The average price at which this order as been executed so far. 0 if the order has not been executed at all.
                      * side - Either "buy" or "sell".
                      * type - Description of the order.
                      * options - An array containing at most one supported order execution option.
                      * timestamp - The timestamp the order was submitted. Note that for compatibility reasons, this is returned as a string. We recommend using the timestampms field instead.
                      * timestampms - The timestamp the order was submitted in milliseconds.
                      * is_live - true if the order is active on the book (has remaining quantity and has not been canceled)
                      * is_cancelled - true if the order has been canceled. Note the spelling, "cancelled" instead of "canceled". This is for compatibility reasons.
                      * reason - Populated with the reason your order was canceled, if available.
                      * was_forced - Will always be false.
                      * executed_amount - The amount of the order that has been filled.
                      * remaining_amount - The amount of the order that has not been filled.
                      * original_amount - The originally submitted amount of the order.
                      * is_hidden - Will always return false unless the order was placed with the indication-of-interest execution option.
    """
    url = URLS.active_orders()
    payload = {
        "request": URLS.get_endpoint(url)
    }
    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err


@login_required
@format_inputs
def order_market(ticker, quantity, side, jsonify=None):
    """ Gemini does not directly support market orders. This function will try to immediately
    place an order or it will cancel it.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param quantity: The amount to trade.
    :type quantity: str
    :param side: Either "buy" or "sell".
    :type side: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * order_id - The order id
                      * client_order_id - An optional client-specified order id
                      * symbol - The symbol of the order
                      * exchange - Will always be "gemini"
                      * price - The price the order was issued at
                      * avg_execution_price - The average price at which this order as been executed so far. 0 if the order has not been executed at all.
                      * side - Either "buy" or "sell".
                      * type - Description of the order.
                      * options - An array containing at most one supported order execution option.
                      * timestamp - The timestamp the order was submitted. Note that for compatibility reasons, this is returned as a string. We recommend using the timestampms field instead.
                      * timestampms - The timestamp the order was submitted in milliseconds.
                      * is_live - true if the order is active on the book (has remaining quantity and has not been canceled)
                      * is_cancelled - true if the order has been canceled. Note the spelling, "cancelled" instead of "canceled". This is for compatibility reasons.
                      * reason - Populated with the reason your order was canceled, if available.
                      * was_forced - Will always be false.
                      * executed_amount - The amount of the order that has been filled.
                      * remaining_amount - The amount of the order that has not been filled.
                      * original_amount - The originally submitted amount of the order.
                      * is_hidden - Will always return false unless the order was placed with the indication-of-interest execution option.
    """
    if side == "buy":
        far_limit_price = float(get_price(ticker, side)) * 10
    else:
        far_limit_price = float(get_price(ticker, side)) / 10
    price = str(round(far_limit_price, 2))
    return order(ticker, quantity, side, price, None, None, ["immediate-or-cancel"], jsonify=jsonify)


@login_required
@format_inputs
def order(ticker, quantity, side, price=None, stop_limit_price=None, min_amount=None, options=None, jsonify=None):
    """ A generic order that can be used for any cryptocurrency.

    :param ticker: The ticker of the crypto.
    :type ticker: str
    :param quantity: The amount to trade.
    :type quantity: str
    :param side: Either "buy" or "sell".
    :type side: str
    :param price: Set this value to set a limit price.
    :type price: Optional[str]
    :param stop_limit_price: Set this value to set a stop price.
    :type stop_limit_price: Optional[str]
    :param min_amount: Minimum decimal amount to purchase, for block trades only.
    :type min_amount: Optional[str]
    :param options: An optional array containing at most one supported order execution option.
    :type options: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error. \
        The keys for the dictionary are listed below.
    :Dictionary Keys: * order_id - The order id
                      * client_order_id - An optional client-specified order id
                      * symbol - The symbol of the order
                      * exchange - Will always be "gemini"
                      * price - The price the order was issued at
                      * avg_execution_price - The average price at which this order as been executed so far. 0 if the order has not been executed at all.
                      * side - Either "buy" or "sell".
                      * type - Description of the order.
                      * options - An array containing at most one supported order execution option.
                      * timestamp - The timestamp the order was submitted. Note that for compatibility reasons, this is returned as a string. We recommend using the timestampms field instead.
                      * timestampms - The timestamp the order was submitted in milliseconds.
                      * is_live - true if the order is active on the book (has remaining quantity and has not been canceled)
                      * is_cancelled - true if the order has been canceled. Note the spelling, "cancelled" instead of "canceled". This is for compatibility reasons.
                      * reason - Populated with the reason your order was canceled, if available.
                      * was_forced - Will always be false.
                      * executed_amount - The amount of the order that has been filled.
                      * remaining_amount - The amount of the order that has not been filled.
                      * original_amount - The originally submitted amount of the order.
                      * is_hidden - Will always return false unless the order was placed with the indication-of-interest execution option.
    """
    url = URLS.order_new()
    payload = {
        "client_order_id": generate_order_id(),
        "request": URLS.get_endpoint(url),
        "symbol": ticker,
        "amount": str(quantity),
        "side": side
    }
    #
    if price:
        payload["price"] = price
    else:
        payload["price"] = get_price(ticker, side)
    #
    if stop_limit_price:
        payload["type"] = "exchange stop limit"
        payload["stop_price"] = stop_limit_price
    else:
        payload["type"] = "exchange limit"
    #
    if min_amount:
        payload["min_amount"] = min_amount
    #
    if options:
        payload["options"] = options

    generate_signature(payload)
    data, err = request_post(url, payload, jsonify)
    return data, err



""" Module contains all the API endpoints """
from enum import Enum, auto
from re import IGNORECASE, match, split

from robin_stocks.gemini.helper import get_sandbox_flag


class AutoName(Enum):
    """Automatically sets an enum value to be its name when using auto()"""

    def _generate_next_value_(name, start, count, last_values):
        return name


class Version(AutoName):
    """Enum for different version types"""
    v1 = auto()
    v2 = auto()


class URLS:
    """ Static class for holding all urls."""
    __base_url = "https://api.tdameritrade.com"

    def __init__(self):
        raise NotImplementedError(
            "Cannot create instance of {0}".format(self.__class__.__name__))

    @classmethod
    def get_base_url(cls, version):
        return cls.__base_url + "/" + version.value + "/"

    @classmethod
    def get_endpoint(cls, url):
        if match(cls.__base_url, url, IGNORECASE):
            _, end = split(cls.__base_url, url, IGNORECASE)
        else:
            raise ValueError("The URL has the wrong base.")

        return end

    # accounts.py
    @classmethod
    def account(cls, id):
        return cls.get_base_url(Version.v1) + "accounts/{0}".format(id)

    @classmethod
    def accounts(cls):
        return cls.get_base_url(Version.v1) + "accounts"

    @classmethod
    def transaction(cls, id, transaction):
        return cls.get_base_url(Version.v1) + "accounts/{0}/transactions/{1}".format(id, transaction)

    @classmethod
    def transactions(cls, id):
        return cls.get_base_url(Version.v1) + "accounts/{0}/transactions".format(id)

    # authentication.py
    @classmethod
    def oauth(cls):
        return cls.get_base_url(Version.v1) + "oauth2/token"

    # markets.py
    @classmethod
    def markets(cls):
        return cls.get_base_url(Version.v1) + "marketdata/hours"

    @classmethod
    def market(cls, market):
        return cls.get_base_url(Version.v1) + "marketdata/{0}/hours".format(market)

    @classmethod
    def movers(cls, index):
        return cls.get_base_url(Version.v1) + "marketdata/{0}/movers".format(index)

    # orders.py
    @classmethod
    def orders(cls, account_id):
        return cls.get_base_url(Version.v1) + "accounts/{0}/orders".format(account_id)

    @classmethod
    def order(cls, account_id, order_id):
        return cls.get_base_url(Version.v1) + "accounts/{0}/orders/{1}".format(account_id, order_id)

    # stocks.py
    @classmethod
    def instruments(cls):
        return cls.get_base_url(Version.v1) + "instruments"

    @classmethod
    def instrument(cls, cusip):
        return cls.get_base_url(Version.v1) + "instruments/{0}".format(cusip)

    @classmethod
    def quote(cls, ticker):
        return cls.get_base_url(Version.v1) + "marketdata/{0}/quotes".format(ticker)

    @classmethod
    def quotes(cls):
        return cls.get_base_url(Version.v1) + "marketdata/quotes"

    @classmethod
    def price_history(cls, ticker):
        return cls.get_base_url(Version.v1) + "marketdata/{0}/pricehistory".format(ticker)

    @classmethod
    def option_chains(cls):
        return cls.get_base_url(Version.v1) + "marketdata/chains"



"""Contains functions for getting information about options."""
import sys
from robin_stocks.robinhood.helper import *
from robin_stocks.robinhood.urls import *

def spinning_cursor():
    """ This is a generator function to yield a character. """
    while True:
        for cursor in '|/-\\':
            yield cursor

spinner = spinning_cursor()

def write_spinner():
    """ Function to create a spinning cursor to tell user that the code is working on getting market data. """
    if get_output()==sys.stdout:
        marketString = 'Loading Market Data '
        sys.stdout.write(marketString)
        sys.stdout.write(next(spinner))
        sys.stdout.flush()
        sys.stdout.write('\b'*(len(marketString)+1))

@login_required
def get_aggregate_positions(info=None, account_number=None):
    """Collapses all option orders for a stock into a single dictionary.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = aggregate_url(account_number=account_number)
    data = request_get(url, 'pagination')
    return(filter_data(data, info))

@login_required
def get_aggregate_open_positions(info=None, account_number=None):
    """Collapses all open option positions for a stock into a single dictionary.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each order. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = aggregate_url(account_number=account_number)
    payload = {'nonzero': 'True'}
    data = request_get(url, 'pagination', payload)
    return(filter_data(data, info))


@login_required
def get_market_options(info=None):
    """Returns a list of all options.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each option. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = option_orders_url()
    data = request_get(url, 'pagination')

    return(filter_data(data, info))


@login_required
def get_all_option_positions(info=None, account_number=None):
    """Returns all option positions ever held for the account.

    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each option. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = option_positions_url(account_number=account_number)
    data = request_get(url, 'pagination')
    return(filter_data(data, info))


@login_required
def get_open_option_positions(account_number=None, info=None):
    """Returns all open option positions for the account.
    
    :param acccount_number: the robinhood account number.
    :type acccount_number: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for each option. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    url = option_positions_url(account_number=account_number)
    payload = {'nonzero': 'True'}
    data = request_get(url, 'pagination', payload)

    return(filter_data(data, info))


def get_chains(symbol, info=None):
    """Returns the chain information of an option.

    :param symbol: The ticker of the stock.
    :type symbol: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the option. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return None

    url = chains_url(symbol)
    data = request_get(url)

    return(filter_data(data, info))

@login_required
def find_tradable_options(symbol, expirationDate=None, strikePrice=None, optionType=None, info=None):
    """Returns a list of all available options for a stock.

    :param symbol: The ticker of the stock.
    :type symbol: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param strikePrice: Represents the strike price of the option.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put' or left blank to get both.
    :type optionType: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for all calls of the stock. If info parameter is provided, \
    a list of strings is returned where the strings are the value of the key that matches info.

    """
    try:
        symbol = symbol.upper().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    url = option_instruments_url()
    if not id_for_chain(symbol):
        print("Symbol {} is not valid for finding options.".format(symbol), file=get_output())
        return [None]

    payload = {'chain_id': id_for_chain(symbol),
               'chain_symbol': symbol,
               'state': 'active'}

    if expirationDate:
        payload['expiration_dates'] = expirationDate
    if strikePrice:
        payload['strike_price'] = strikePrice
    if optionType:
        payload['type'] = optionType

    data = request_get(url, 'pagination', payload)
    return(filter_data(data, info))

@login_required
def find_options_by_expiration(inputSymbols, expirationDate, optionType=None, info=None):
    """Returns a list of all the option orders that match the seach parameters

    :param inputSymbols: The ticker of either a single stock or a list of stocks.
    :type inputSymbols: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param optionType: Can be either 'call' or 'put' or leave blank to get both.
    :type optionType: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for all options of the stock that match the search parameters. \
    If info parameter is provided, a list of strings is returned where the strings are the value of the key that matches info.

    """
    try:
        symbols = inputs_to_set(inputSymbols)
        if optionType:
            optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    data = []
    for symbol in symbols:
        allOptions = find_tradable_options(symbol, expirationDate, None, optionType, None)
        filteredOptions = [item for item in allOptions if item.get("expiration_date") == expirationDate]

        for item in filteredOptions:
            marketData = get_option_market_data_by_id(item['id'])
            if marketData:
                item.update(marketData[0])
            write_spinner()

        data.extend(filteredOptions)

    return(filter_data(data, info))

@login_required
def find_options_by_strike(inputSymbols, strikePrice, optionType=None, info=None):
    """Returns a list of all the option orders that match the seach parameters

    :param inputSymbols: The ticker of either a single stock or a list of stocks.
    :type inputSymbols: str
    :param strikePrice: Represents the strike price to filter for.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put' or leave blank to get both.
    :type optionType: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for all options of the stock that match the search parameters. \
    If info parameter is provided, a list of strings is returned where the strings are the value of the key that matches info.

    """
    try:
        symbols = inputs_to_set(inputSymbols)
        if optionType:
            optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    data = []
    for symbol in symbols:
        filteredOptions = find_tradable_options(symbol, None, strikePrice, optionType, None)

        for item in filteredOptions:
            marketData = get_option_market_data_by_id(item['id'])
            if marketData:
                item.update(marketData[0])
            write_spinner()

        data.extend(filteredOptions)

    return(filter_data(data, info))

@login_required
def find_options_by_expiration_and_strike(inputSymbols, expirationDate, strikePrice, optionType=None, info=None):
    """Returns a list of all the option orders that match the seach parameters

    :param inputSymbols: The ticker of either a single stock or a list of stocks.
    :type inputSymbols: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param strikePrice: Represents the strike price to filter for.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put' or leave blank to get both.
    :type optionType: Optional[str]
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for all options of the stock that match the search parameters. \
    If info parameter is provided, a list of strings is returned where the strings are the value of the key that matches info.

    """
    try:
        symbols = inputs_to_set(inputSymbols)
        if optionType:
            optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    data = []
    for symbol in symbols:
        allOptions = find_tradable_options(symbol, expirationDate, strikePrice, optionType, None)
        filteredOptions = [item for item in allOptions if item.get("expiration_date") == expirationDate]

        for item in filteredOptions:
            marketData = get_option_market_data_by_id(item['id'])
            if marketData:
                item.update(marketData[0])
            write_spinner()

        data.extend(filteredOptions)

    return filter_data(data, info)

@login_required
def find_options_by_specific_profitability(inputSymbols, expirationDate=None, strikePrice=None, optionType=None, typeProfit="chance_of_profit_short", profitFloor=0.0, profitCeiling=1.0, info=None):
    """Returns a list of option market data for several stock tickers that match a range of profitability.

    :param inputSymbols: May be a single stock ticker or a list of stock tickers.
    :type inputSymbols: str or list
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD. Leave as None to get all available dates.
    :type expirationDate: str
    :param strikePrice: Represents the price of the option. Leave as None to get all available strike prices.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put' or leave blank to get both.
    :type optionType: Optional[str]
    :param typeProfit: Will either be "chance_of_profit_short" or "chance_of_profit_long".
    :type typeProfit: str
    :param profitFloor: The lower percentage on scale 0 to 1.
    :type profitFloor: int
    :param profitCeiling: The higher percentage on scale 0 to 1.
    :type profitCeiling: int
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a list of dictionaries of key/value pairs for all stock option market data. \
    If info parameter is provided, a list of strings is returned where the strings are the value of the key that matches info.

    """
    symbols = inputs_to_set(inputSymbols)
    data = []

    if (typeProfit != "chance_of_profit_short" and typeProfit != "chance_of_profit_long"):
        print("Invalid string for 'typeProfit'. Defaulting to 'chance_of_profit_short'.", file=get_output())
        typeProfit = "chance_of_profit_short"

    for symbol in symbols:
        tempData = find_tradable_options(symbol, expirationDate, strikePrice, optionType, info=None)
        for option in tempData:
            if expirationDate and option.get("expiration_date") != expirationDate:
                continue

            market_data = get_option_market_data_by_id(option['id'])
            
            if len(market_data):
                option.update(market_data[0])
                write_spinner()

                try:
                    floatValue = float(option[typeProfit])
                    if (floatValue >= profitFloor and floatValue <= profitCeiling):
                        data.append(option)
                except:
                    pass

    return(filter_data(data, info))

@login_required
def get_option_market_data_by_id(id, info=None):
    """Returns the option market data for a stock, including the greeks,
    open interest, change of profit, and adjusted mark price.

    :param id: The id of the stock.
    :type id: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the stock. \
    If info parameter is provided, the value of the key that matches info is extracted.

    """
    instrument = get_option_instrument_data_by_id(id)
    if instrument is None:
      # e.g. 503 Server Error: Service Unavailable for url: https://api.robinhood.com/options/instruments/d1058013-09a2-4063-b6b0-92717e17d0c0/
      return None  # just return None which the caller can easily check; do NOT use faked empty data, it will only cause future problem
    else:
      payload = {
          "instruments" : instrument['url']
      }
      url = marketdata_options_url()
      data = request_get(url, 'results', payload)

    return(filter_data(data, info))

@login_required
def get_option_market_data(inputSymbols, expirationDate, strikePrice, optionType, info=None):
    """Returns the option market data for the stock option, including the greeks,
    open interest, change of profit, and adjusted mark price.

    :param inputSymbols: The ticker of the stock.
    :type inputSymbols: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param strikePrice: Represents the price of the option.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put'.
    :type optionType: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the stock. \
    If info parameter is provided, the value of the key that matches info is extracted.

    """
    try:
        symbols = inputs_to_set(inputSymbols)
        if optionType:
            optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    data = []
    for symbol in symbols:
        optionID = id_for_option(symbol, expirationDate, strikePrice, optionType)
        marketData = get_option_market_data_by_id(optionID)
        data.append(marketData)

    return(filter_data(data, info))


def get_option_instrument_data_by_id(id, info=None):
    """Returns the option instrument information.

    :param id: The id of the stock.
    :type id: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the stock. \
    If info parameter is provided, the value of the key that matches info is extracted.

    """
    url = option_instruments_url(id)
    data = request_get(url)
    return(filter_data(data, info))


def get_option_instrument_data(symbol, expirationDate, strikePrice, optionType, info=None):
    """Returns the option instrument data for the stock option.

    :param symbol: The ticker of the stock.
    :type symbol: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param strikePrice: Represents the price of the option.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put'.
    :type optionType: str
    :param info: Will filter the results to get a specific value.
    :type info: Optional[str]
    :returns: Returns a dictionary of key/value pairs for the stock. \
    If info parameter is provided, the value of the key that matches info is extracted.

    """
    try:
        symbol = symbol.upper().strip()
        optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    optionID = id_for_option(symbol, expirationDate, strikePrice, optionType)
    url = option_instruments_url(optionID)
    data = request_get(url)

    return(filter_data(data, info))


def get_option_historicals(symbol, expirationDate, strikePrice, optionType, interval='hour', span='week', bounds='regular', info=None):
    """Returns the data that is used to make the graphs.

    :param symbol: The ticker of the stock.
    :type symbol: str
    :param expirationDate: Represents the expiration date in the format YYYY-MM-DD.
    :type expirationDate: str
    :param strikePrice: Represents the price of the option.
    :type strikePrice: str
    :param optionType: Can be either 'call' or 'put'.
    :type optionType: str
    :param interval: Interval to retrieve data for. Values are '5minute', '10minute', 'hour', 'day', 'week'. Default is 'hour'.
    :type interval: Optional[str]
    :param span: Sets the range of the data to be either 'day', 'week', 'year', or '5year'. Default is 'week'.
    :type span: Optional[str]
    :param bounds: Represents if graph will include extended trading hours or just regular trading hours. Values are 'regular', 'trading', and 'extended'. \
    regular hours are 6 hours long, trading hours are 9 hours long, and extended hours are 16 hours long. Default is 'regular'
    :type bounds: Optional[str]
    :param info: Will filter the results to have a list of the values that correspond to key that matches info.
    :type info: Optional[str]
    :returns: Returns a list that contains a list for each symbol. \
    Each list contains a dictionary where each dictionary is for a different time.

    """
    try:
        symbol = symbol.upper().strip()
        optionType = optionType.lower().strip()
    except AttributeError as message:
        print(message, file=get_output())
        return [None]

    interval_check = ['5minute', '10minute', 'hour', 'day', 'week']
    span_check = ['day', 'week', 'year', '5year']
    bounds_check = ['extended', 'regular', 'trading']
    if interval not in interval_check:
        print(
            'ERROR: Interval must be "5minute","10minute","hour","day",or "week"', file=get_output())
        return([None])
    if span not in span_check:
        print('ERROR: Span must be "day", "week", "year", or "5year"', file=get_output())
        return([None])
    if bounds not in bounds_check:
        print('ERROR: Bounds must be "extended","regular",or "trading"', file=get_output())
        return([None])

    optionID = id_for_option(symbol, expirationDate, strikePrice, optionType)

    url = option_historicals_url(optionID)
    payload = {'span': span,
               'interval': interval,
               'bounds': bounds}
    data = request_get(url, 'regular', payload)
    if (data == None or data == [None]):
        return data

    histData = []
    for subitem in data['data_points']:
        subitem['symbol'] = symbol
        histData.append(subitem)

    return(filter_data(histData, info))



from .accounts import (get_account, get_accounts, get_transaction,
                       get_transactions)
from .authentication import (generate_encryption_passcode, login,
                             login_first_time)
from .helper import (get_login_state, get_order_number, request_data,
                     request_delete, request_get, request_headers,
                     request_post)
from .markets import get_hours_for_market, get_hours_for_markets, get_movers
from .orders import (cancel_order, get_order, get_orders_for_account,
                     place_order)
from .stocks import (get_instrument, get_option_chains, get_price_history,
                     get_quote, get_quotes, search_instruments)



"""Holds the session header and other global variables."""
from requests import Session

NONCE = 1 # Counter that must always be increasing
LOGGED_IN = False # Flag on whether or not the user is logged in.
USE_SANDBOX_URLS = False # Flag on whether or not to use sandbox urls.
RETURN_PARSED_JSON_RESPONSE = False # Flag on whether to automatically parse request responses.
SECRET_API_KEY = None

# The session object for making get and post requests.
SESSION = Session()
SESSION.headers = {
    'Content-Type': "text/plain",
    'Content-Length': "0",
    'Cache-Control': "no-cache"
}


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='robin_stocks',
      version='3.3.0',
      description='A Python wrapper around the Robinhood API',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      url='https://github.com/jmfernandes/robin_stocks',
      author='Josh Fernandes',
      author_email='joshfernandes@mac.com',
      keywords=['robinhood','robin stocks','finance app','stocks','options','trading','investing'],
      license='MIT',
      python_requires='>=3.9',
      packages=find_packages(),
      requires=['requests', 'pyotp', 'cryptography'],
      install_requires=[
          'requests',
          'pyotp',
          'python-dotenv',
          'cryptography'
      ],
      zip_safe=False)



from robin_stocks.tda.helper import (format_inputs, login_required,
                                     request_delete, request_get,
                                     request_headers)
from robin_stocks.tda.urls import URLS


@login_required
@format_inputs
def place_order(account_id, order_payload, jsonify=None):
    """ Place an order for a given account.

    :param account_id: The account id.
    :type account_id: str
    :param order_payload: A dictionary of key value pairs for the infromation you want to send to order.
    :type order_payload: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.orders(account_id)
    data, error = request_headers(url, order_payload, jsonify)
    return data, error


@login_required
@format_inputs
def cancel_order(account_id, order_id, jsonify=None):
    """ Cancel an order for a given account.

    :param account_id: The account id.
    :type account_id: str
    :param order_id: The order id.
    :type order_id: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.order(account_id, order_id)
    data, error = request_delete(url, jsonify)
    return data, error


@login_required
@format_inputs
def get_order(account_id, order_id, jsonify=None):
    """ Gets information for an order for a given account.

    :param account_id: The account id.
    :type account_id: str
    :param order_id: The order id.
    :type order_id: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.order(account_id, order_id)
    data, error = request_get(url, None, jsonify)
    return data, error


@login_required
@format_inputs
def get_orders_for_account(account_id, max_results=None, from_time=None, to_time=None, status=None, jsonify=None):
    """ Gets all the orders for a given account.

    :param account_id: The account id.
    :type account_id: Optional[str]
    :param max_results: The max number of orders to retrieve.
    :type max_results: Optional[str]
    :param from_time: Specifies that no orders entered before this time should be returned. Valid ISO-8601 formats are : \
yyyy-MM-dd. Date must be within 60 days from today's date. 'toEnteredTime' must also be set.
    :type from_time: Optional[str]
    :param to_time: Specifies that no orders entered after this time should be returned.Valid ISO-8601 formats are : \
yyyy-MM-dd. 'fromEnteredTime' must also be set.
    :type to_time: Optional[str]
    :param status: Specifies that only orders of this status should be returned. Possible values are \
        AWAITING_PARENT_ORDER, AWAITING_CONDITION, AWAITING_MANUAL_REVIEW, ACCEPTED, AWAITING_UR_OUT, PENDING_ACTIVATION, QUEUED \
        WORKING, REJECTED, PENDING_CANCEL, CANCELED, PENDING_REPLACE, REPLACED, FILLED, EXPIRED
    :type status: Optional[str]
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.orders(account_id)
    payload = {}
    if max_results:
        payload["maxResults"] = max_results
    if from_time:
        payload["fromEnteredTime"] = from_time
    if to_time:
        payload["toEnteredTime"] = to_time
    if status:
        payload["status"] = status
    data, error = request_get(url, payload, jsonify)
    return data, error



from base64 import urlsafe_b64decode as b64d
from base64 import urlsafe_b64encode as b64e
from functools import wraps
from inspect import signature
from zlib import compress, decompress

from robin_stocks.gemini.globals import (LOGGED_IN, NONCE,
                                         RETURN_PARSED_JSON_RESPONSE,
                                         SECRET_API_KEY, SESSION,
                                         USE_SANDBOX_URLS)


def increment_nonce():
    """ Increase nonce by one.
    """
    global NONCE
    NONCE += 1


def get_nonce():
    """ return the nonce.
    """
    return NONCE


def set_secret_key(data):
    """ Encodes the secret api key before storing it as a global variable.
    """
    global SECRET_API_KEY
    SECRET_API_KEY = b64e(compress(data, 9))


def get_secret_key():
    """ Decodes the secret api key from the global variable.
    """
    return decompress(b64d(SECRET_API_KEY))


def format_inputs(func):
    """ A decorator for formatting inputs. For any function decorated by this,
        the value of jsonify=None will be replaced with the global value stored at 
        RETURN_PARSED_JSON_RESPONSE.
    """
    @wraps(func)
    def format_wrapper(*args, **kwargs):
        bound_args = signature(func).bind(*args, **kwargs)
        bound_args.apply_defaults()
        target_args = dict(bound_args.arguments)
        if target_args['jsonify'] is None:
            kwargs["jsonify"] = get_default_json_flag()
        return(func(*args, **kwargs))
    return(format_wrapper)


def set_default_json_flag(parse_json):
    """ Sets whether you want all functions to return the json parsed response or not.

    :param parse_json: Set to change value of global variable.
    :type parse_json: bool
    """
    global RETURN_PARSED_JSON_RESPONSE
    RETURN_PARSED_JSON_RESPONSE = parse_json


def get_default_json_flag():
    """ Gets the boolean flag on the default JSON setting.
    """
    return RETURN_PARSED_JSON_RESPONSE


def get_sandbox_flag():
    """ Gets whether all functions are to use the sandbox base url.
    """
    return USE_SANDBOX_URLS


def use_sand_box_urls(use_sandbox):
    """ Sets whether you want all functions to use the sandbox base url.

    :param use_sandbox: Set to change value of global variable.
    :type use_sandbox: bool
    """
    global USE_SANDBOX_URLS
    USE_SANDBOX_URLS = use_sandbox


def update_session(key, value):
    """Updates the session header used by the requests library.

    :param key: The key value to update or add to session header.
    :type key: str
    :param value: The value that corresponds to the key.
    :type value: str

    """
    SESSION.headers[key] = value


def set_login_state(logged_in):
    """ Sets the login state

    :param logged_in: Set to change value of global variable.
    :type logged_in: bool
    """
    global LOGGED_IN
    LOGGED_IN = logged_in


def get_login_state():
    """ Gets the login state
    """
    return LOGGED_IN


def login_required(func):
    """ A decorator for indicating which methods require the user to be logged in.
    """
    @wraps(func)
    def login_wrapper(*args, **kwargs):
        global LOGGED_IN
        if not LOGGED_IN:
            raise Exception('{} can only be called when logged in'.format(
                func.__name__))
        return(func(*args, **kwargs))
    return(login_wrapper)


def request_get(url, payload, parse_json):
    """ Generic function for sending a get request.

    :param url: The url to send a get request to.
    :type url: str
    :param payload: Dictionary of parameters to pass to the url. Will append the requests url as url/?key1=value1&key2=value2.
    :type payload: dict
    :param parse_json: Requests serializes data in the JSON format. Set this parameter true to parse the data to a dictionary \
        using the JSON format.
    :type parse_json: bool
    :returns: Returns a tuple where the first entry is the response and the second entry will be an error message from the \
        get request. If there was no error then the second entry in the tuple will be None. The first entry will either be \
        the raw request response or the parsed JSON response based on whether parse_json is True or not.
    """
    response_error = None
    try:
        response = SESSION.get(url, params=payload)
        response.raise_for_status()
    except Exception as e:
        response_error = e
    # Return either the raw request object so you can call response.text, response.status_code, response.headers, or response.json()
    # or return the JSON parsed information if you don't care to check the status codes.
    if parse_json:
        return response.json(), response_error
    else:
        return response, response_error


def request_post(url, payload, parse_json):
    """ Generic function for sending a post request.

    :param url: The url to send a post request to.
    :type url: str
    :param payload: Dictionary of parameters to pass to the url. Will append the requests url as url/?key1=value1&key2=value2.
    :type payload: dict
    :param parse_json: Requests serializes data in the JSON format. Set this parameter true to parse the data to a dictionary \
        using the JSON format.
    :type parse_json: bool
    :returns: Returns a tuple where the first entry is the response and the second entry will be an error message from the \
        get request. If there was no error then the second entry in the tuple will be None. The first entry will either be \
        the raw request response or the parsed JSON response based on whether parse_json is True or not.
    """
    response_error = None
    try:
        response = SESSION.post(url, params=payload)
        response.raise_for_status()
    except Exception as e:
        response_error = e
    # Return either the raw request object so you can call response.text, response.status_code, response.headers, or response.json()
    # or return the JSON parsed information if you don't care to check the status codes.
    if parse_json:
        return response.json(), response_error
    else:
        return response, response_error



from robin_stocks.tda.helper import format_inputs, login_required, request_get
from robin_stocks.tda.urls import URLS


@login_required
@format_inputs
def get_hours_for_markets(markets, date, jsonify=None):
    """ Gets market hours for various markets.

    :param markets: The markets for which you're requesting market hours, comma-separated. \
        Valid markets are EQUITY, OPTION, FUTURE, BOND, or FOREX.
    :type markets: str
    :param date: The date for which market hours information is requested. Valid ISO-8601 formats are : \
        yyyy-MM-dd and yyyy-MM-dd'T'HH:mm:ssz.
    :type date: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.markets()
    payload = {
        "markets": markets,
        "date": date
    }
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_hours_for_market(market, date, jsonify=None):
    """ Gets market hours for a specific market.

    :param market: The market for which you're requesting market hours, comma-separated. \
        Valid markets are EQUITY, OPTION, FUTURE, BOND, or FOREX.
    :type market: str
    :param date: The date for which market hours information is requested. Valid ISO-8601 formats are : \
        yyyy-MM-dd and yyyy-MM-dd'T'HH:mm:ssz.
    :type date: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.market(market)
    payload = {
        "date": date
    }
    data, error = request_get(url, payload, jsonify)
    return data, error


@login_required
@format_inputs
def get_movers(market, direction, change, jsonify=None):
    """ Gets market hours for a specific market.

    :param market: The market for which you're requesting market hours, comma-separated. \
        Valid markets are $DJI, $COMPX, or $SPX.X.
    :type market: str
    :param direction: To return movers with the specified directions of "up" or "down".
    :type direction: str
    :param change: To return movers with the specified change types of "percent" or "value".
    :type change: str
    :param jsonify: If set to false, will return the raw response object. \
        If set to True, will return a dictionary parsed using the JSON format.
    :type jsonify: Optional[str]
    :returns: Returns a tuple where the first entry in the tuple is a requests reponse object  \
        or a dictionary parsed using the JSON format and the second entry is an error string or \
        None if there was not an error.

    """
    url = URLS.movers(market)
    payload = {
        "direction": direction,
        "change": change
    }
    data, error = request_get(url, payload, jsonify)
    return data, error
